import importlib
import socket
import time
import sys
from contextlib import contextmanager
from docassemble.base.config import daconfig
from docassemble.base.logger import logmessage
from celery import Celery
from celery.result import result_from_tuple

backend = daconfig.get('redis', None)
if backend is None:
    backend = 'redis://localhost'
broker = daconfig.get('rabbitmq', None)
if broker is None:
    broker = 'pyamqp://guest@' + socket.gethostname() + '//'

workerapp = Celery('docassemble.webapp.worker', backend=backend, broker=broker)
importlib.import_module('docassemble.webapp.config_worker')
workerapp.config_from_object('docassemble.webapp.config_worker')
workerapp.set_current()
workerapp.set_default()


class WorkerController:

    def __init__(self):
        self.loaded = False

    def initialize(self):
        if self.loaded:
            return
        from docassemble.webapp.server import set_request_active, fetch_user_dict, save_user_dict, obtain_lock, obtain_lock_patiently, release_lock, Message, reset_user_dict, da_send_mail, get_info_from_file_number, retrieve_email, trigger_update, r, apiclient, get_ext_and_mimetype, get_user_object, login_user, error_notification, noquote, update_last_login  # pylint: disable=import-outside-toplevel
        from docassemble.webapp.server import app as flaskapp  # pylint: disable=import-outside-toplevel
        import docassemble.webapp.update  # pylint: disable=import-outside-toplevel, redefined-outer-name
        import docassemble.base.functions  # pylint: disable=import-outside-toplevel
        docassemble.base.functions.server_context.context = 'celery'
        import docassemble.base.interview_cache  # pylint: disable=import-outside-toplevel
        import docassemble.base.util  # pylint: disable=import-outside-toplevel
        import docassemble.base.parse  # pylint: disable=import-outside-toplevel
        self.flaskapp = flaskapp
        self.set_request_active = set_request_active
        self.fetch_user_dict = fetch_user_dict
        self.save_user_dict = save_user_dict
        self.obtain_lock = obtain_lock
        self.obtain_lock_patiently = obtain_lock_patiently
        self.release_lock = release_lock
        self.Message = Message
        self.reset_user_dict = reset_user_dict
        self.da_send_mail = da_send_mail
        self.functions = docassemble.base.functions
        self.interview_cache = docassemble.base.interview_cache
        self.parse = docassemble.base.parse
        self.retrieve_email = retrieve_email
        self.get_info_from_file_number = get_info_from_file_number
        self.trigger_update = trigger_update
        self.util = docassemble.base.util
        self.r = r
        self.apiclient = apiclient
        self.get_ext_and_mimetype = get_ext_and_mimetype
        self.get_user_object = get_user_object
        self.login_user = login_user
        self.update_last_login = update_last_login
        self.error_notification = error_notification
        self.noquote = noquote
        self.update = docassemble.webapp.update
        self.loaded = True


worker_controller = WorkerController()


def convert(obj):
    return result_from_tuple(obj.as_tuple(), app=workerapp)


def process_error(interview, session_code, yaml_filename, secret, user_info, url, url_root, is_encrypted, error_type, error_message, error_trace, variables, extra):
    start_time = time.time()
    new_action = worker_controller.functions.this_thread.current_info['on_error']
    new_action['arguments']['error_type'] = error_type
    new_action['arguments']['error_message'] = error_message
    new_action['arguments']['error_trace'] = error_trace
    new_action['arguments']['variables'] = variables
    the_current_info = {'user': user_info, 'session': session_code, 'secret': secret, 'yaml_filename': yaml_filename, 'url': url, 'url_root': url_root, 'encrypted': is_encrypted, 'interface': 'worker', 'action': new_action['action'], 'arguments': new_action['arguments']}
    worker_controller.functions.this_thread.current_info = the_current_info
    worker_controller.obtain_lock(session_code, yaml_filename)
    steps, user_dict, is_encrypted = worker_controller.fetch_user_dict(session_code, yaml_filename, secret=secret)
    interview_status = worker_controller.parse.InterviewStatus(current_info=the_current_info)
    old_language = worker_controller.functions.get_language()
    try:
        interview.assemble(user_dict, interview_status)
    except Exception as e:
        if hasattr(e, 'traceback'):
            logmessage("Error in assembly during error callback: " + str(e.__class__.__name__) + ": " + str(e) + ": " + str(e.traceback))
        else:
            logmessage("Error in assembly during error callback: " + str(e.__class__.__name__) + ": " + str(e))
        error_type = e.__class__.__name__
        error_message = str(e)
        if hasattr(e, 'traceback'):
            error_trace = str(e.traceback)
            if hasattr(e, 'da_line_with_error'):
                error_trace += "\nIn line: " + str(e.da_line_with_error)
        else:
            error_trace = None
        worker_controller.error_notification(e, message=error_message, trace=error_trace)
    worker_controller.functions.set_language(old_language)
    # is this right?
    save_status = worker_controller.functions.this_thread.misc.get('save_status', 'new')
    if save_status != 'ignore':
        if str(user_info.get('the_user_id', None)).startswith('t'):
            worker_controller.save_user_dict(session_code, user_dict, yaml_filename, secret=secret, encrypt=is_encrypted, steps=steps)
        else:
            worker_controller.save_user_dict(session_code, user_dict, yaml_filename, secret=secret, encrypt=is_encrypted, manual_user_id=user_info['theid'], steps=steps)
    worker_controller.release_lock(session_code, yaml_filename)
    if hasattr(interview_status, 'question'):
        if interview_status.question.question_type == "response":
            logmessage("Time in error callback was " + str(time.time() - start_time))
            # logmessage("background_action: status in error callback was response")
            if hasattr(interview_status.question, 'all_variables'):
                pass
            elif not hasattr(interview_status.question, 'binaryresponse'):
                sys.stdout.write(interview_status.questionText.rstrip().encode('utf8') + "\n")
        elif interview_status.question.question_type == "backgroundresponse":
            logmessage("Time in error callback was " + str(time.time() - start_time))
            return worker_controller.functions.ReturnValue(ok=False, error_type=error_type, error_trace=error_trace, error_message=error_message, variables=variables, value=interview_status.question.backgroundresponse, extra=extra)
    logmessage("Time in error callback was " + str(time.time() - start_time))
    return worker_controller.functions.ReturnValue(ok=False, error_type=error_type, error_trace=error_trace, error_message=error_message, variables=variables, extra=extra)


def error_object(err):
    logmessage("Error: " + str(err.__class__.__name__) + ": " + str(err))
    error_type = err.__class__.__name__
    error_message = str(err)
    error_trace = None
    worker_controller.error_notification(err, message=error_message, trace=error_trace)
    return worker_controller.functions.ReturnValue(ok=False, error_message=error_message, error_type=error_type, error_trace=error_trace, restart=False)


@contextmanager
def bg_context():
    worker_controller.initialize()
    url_root = daconfig.get('url root', 'http://localhost') + daconfig.get('root', '/')
    url = url_root + 'interview'
    with worker_controller.flaskapp.app_context() as a, worker_controller.flaskapp.test_request_context(base_url=url_root, path=url) as b:
        worker_controller.functions.reset_local_variables()
        worker_controller.set_request_active(False)
        yield (a, b)



from flask_mail import Mail, Connection as FlaskConnection, Message as FlaskMessage


class Connection(FlaskConnection):
    def send(self, message, envelope_from=None):  # pylint: disable=unused-argument
        if message.sender is None:
            message.sender = self.mail.default_sender
        super().send(message, envelope_from=envelope_from)


class FlaskMail(Mail):

    def __init__(self, app=None, config=None):
        self.app = app
        if app is not None or config is not None:
            self.state = self.init_app(app=app, config=config)
        else:
            self.state = None

    def init_app(self, app=None, config=None):
        if app is not None:
            if config is None:
                config = app.config
            state = self.init_mail(config, app.debug, app.testing)
            app.extensions = getattr(app, 'extensions', {})
            app.extensions['mail'] = state
        else:
            state = self.init_mail(config, False, False)
        return state

    def connect(self):
        try:
            return Connection(self.state)
        except KeyError:
            raise RuntimeError("The curent application was not configured with Flask-Mail")


class Message(FlaskMessage):

    def __init__(self, subject='',
                 recipients=None,
                 body=None,
                 html=None,
                 sender=None,
                 cc=None,
                 bcc=None,
                 attachments=None,
                 reply_to=None,
                 date=None,
                 charset=None,
                 extra_headers=None,
                 mail_options=None,
                 rcpt_options=None):
        null_sender = sender is None
        super().__init__(subject, recipients, body, html, sender, cc, bcc, attachments, reply_to, date, charset, extra_headers, mail_options, rcpt_options)
        if null_sender:
            self.sender = None



from random import randint
import json
import redis

r = redis.Redis()
r.publish('mychan', json.dumps({'origin': 'server', 'room': 'RMXgFEzlgqdZsDSSJPPFuzDdoMYuKkBG', 'message': 'Hey random ' + str(randint(0, 99))}))



import importlib
from celery import Celery, chord  # noqa: F401 # pylint: disable=unused-import
import docassemble.base.config
if not docassemble.base.config.loaded:
    docassemble.base.config.load(in_celery=True)
from docassemble.base.config import daconfig
from docassemble.base.logger import logmessage
from docassemble.webapp.worker_common import workerapp, convert, bg_context  # noqa: F401 # pylint: disable=unused-import
from docassemble.webapp.worker_tasks import background_action, ocr_page, ocr_google, make_png_for_pdf, email_attachments, update_packages, sync_with_google_drive, sync_with_onedrive, ocr_dummy, ocr_finalize, reset_server  # noqa: F401 # pylint: disable=unused-import

for module_name in daconfig['celery modules']:
    try:
        importlib.import_module(module_name)
    except:
        logmessage("Error importing " + module_name + " from celery modules list")



import re
from flask_wtf import FlaskForm
from docassemble.base.functions import LazyWord as word
from wtforms import validators, ValidationError, StringField, SubmitField, TextAreaField, SelectMultipleField, SelectField, FileField, HiddenField, RadioField, BooleanField
import packaging


class NonValidatingSelectField(SelectField):

    def pre_validate(self, form):
        pass


def validate_project_name(form, field):  # pylint: disable=unused-argument
    if re.search(r'^[0-9]', field.data):
        raise ValidationError(word('Project name cannot begin with a number'))
    if re.search(r'[^A-Za-z0-9]', field.data):
        raise ValidationError(word('Valid characters are: A-Z, a-z, 0-9'))


def validate_name(form, field):  # pylint: disable=unused-argument
    if re.search(r'[^A-Za-z0-9\-]', field.data):
        raise ValidationError(word('Valid characters are: A-Z, a-z, 0-9, hyphen'))


def validate_package_version(form, field):  # pylint: disable=unused-argument
    try:
        packaging.version.Version(field.data)
    except packaging.version.InvalidVersion:
        raise ValidationError(word('Version number does not conform to PEP 440'))


def validate_package_name(form, field):  # pylint: disable=unused-argument
    if re.search(r'[^A-Za-z0-9]', field.data):
        raise ValidationError(word('Valid characters are: A-Z, a-z, 0-9'))


class CreatePackageForm(FlaskForm):
    name = StringField(word('Package name'), validators=[
        validators.DataRequired(word('Package name is required')), validate_name])
    submit = SubmitField(word('Get template'))


class CreatePlaygroundPackageForm(FlaskForm):
    name = SelectField(word('Package'), validators=[
        validators.DataRequired(word('Package name is required')), validate_name])
    submit = SubmitField(word('Get package'))


class UpdatePackageForm(FlaskForm):
    giturl = StringField(word('GitHub URL'))
    gitbranch = SelectField(word('GitHub Branch'))
    zipfile = FileField(word('Zip File'))
    pippackage = StringField(word('Package on PyPI'))
    submit = SubmitField(word('Update'))


class ConfigForm(FlaskForm):
    config_content = TextAreaField(word('Configuration YAML'))
    submit = SubmitField(word('Save'))
    cancel = SubmitField(word('Cancel'))


class PlaygroundForm(FlaskForm):
    status = StringField('Status')
    original_playground_name = StringField(word('Original Name'))
    playground_name = StringField(word('Name'), [validators.Length(min=1, max=255)])
    playground_content = TextAreaField(word('Playground YAML'))
    search_term = StringField(word('Search'))
    submit = SubmitField(word('Save'))
    run = SubmitField(word('Save and Run'))
    delete = SubmitField(word('Delete'))


class PlaygroundUploadForm(FlaskForm):
    uploadfile = FileField(word('File to upload'))


class LogForm(FlaskForm):
    filter_string = StringField(word('Filter For'))
    file_name = StringField(word('File Name'))
    submit = SubmitField(word('Apply'))
    clear = SubmitField(word('Clear'))


class Utilities(FlaskForm):
    pdfdocxfile = FileField(word('PDF/DOCX File'))
    scan = SubmitField(word('Scan'))
    interview = StringField(word('Interview'))
    interview_submit = SubmitField(word('Download'))
    language = StringField(word('Language'))
    tr_language = StringField(word('Language'))
    systemfiletype = SelectField(word('Output Format'))
    filetype = SelectField(word('Output Format'))
    language_submit = SubmitField(word('Translate'))
    officeaddin_version = StringField(word('Version'), default='0.0.0.1')
    officeaddin_submit = SubmitField(word('Download'))


class PlaygroundFilesForm(FlaskForm):
    purpose = StringField('Purpose')
    section = StringField(word('Section'))
    uploadfile = FileField(word('File to upload'))
    submit = SubmitField(word('Upload'))


class PlaygroundFilesEditForm(FlaskForm):
    purpose = StringField('Purpose')
    section = StringField(word('Section'))
    original_file_name = StringField(word('Original Name'))
    file_name = StringField(word('Name'), [validators.Length(min=1, max=255)])
    search_term = StringField(word('Search'))
    file_content = TextAreaField(word('File Text'))
    active_file = StringField(word('Active File'))
    submit = SubmitField(word('Save'))
    delete = SubmitField(word('Delete'))


class RenameProject(FlaskForm):
    name = StringField(word('New Name'), validators=[
        validators.DataRequired(word('Project name is required')), validate_project_name])
    submit = SubmitField(word('Rename'))


class DeleteProject(FlaskForm):
    submit = SubmitField(word('Delete'))


class NewProject(FlaskForm):
    name = StringField(word('Name'), validators=[
        validators.DataRequired(word('Project name is required')), validate_project_name])
    submit = SubmitField(word('Save'))


class PullPlaygroundPackage(FlaskForm):
    github_url = StringField(word('GitHub URL'))
    github_branch = SelectField(word('GitHub Branch'))
    pypi = StringField(word('PyPI package'))
    pull = SubmitField(word('Pull'))
    cancel = SubmitField(word('Cancel'))


class PlaygroundPackagesForm(FlaskForm):
    original_file_name = StringField(word('Original Name'))
    file_name = StringField(word('Package Name'), validators=[validators.Length(min=1, max=50),
                                                              validators.DataRequired(word('Package Name is required')),
                                                              validate_package_name])
    license = StringField(word('License'), default='The MIT License (MIT)', validators=[validators.Length(min=0, max=255)])
    author_name = StringField(word('Author Name'), validators=[validators.Length(min=0, max=255)])
    author_email = StringField(word('Author E-mail'), validators=[validators.Length(min=0, max=255)])
    description = StringField(word('Description'), validators=[validators.Length(min=0, max=255)], default="A docassemble extension.")
    version = StringField(word('Version'), validators=[validators.Length(min=0, max=255), validate_package_version], default="0.0.1")
    url = StringField(word('URL'), validators=[validators.Length(min=0, max=255)], default="")
    dependencies = SelectMultipleField(word('Dependencies'))
    interview_files = SelectMultipleField(word('Interview files'))
    template_files = SelectMultipleField(word('Template files'))
    module_files = SelectMultipleField(word('Modules'))
    static_files = SelectMultipleField(word('Static files'))
    sources_files = SelectMultipleField(word('Source files'))
    readme = TextAreaField(word('README file'), default='')
    github_branch = NonValidatingSelectField(word('Branch'))
    github_branch_new = StringField(word('Name of new branch'))
    commit_message = StringField(word('Commit message'), default="")
    pypi_also = BooleanField(word('Publish on PyPI also'))
    install_also = BooleanField(word('Install package on this server also'))
    submit = SubmitField(word('Save'))
    download = SubmitField(word('Download'))
    install = SubmitField(word('Install'))
    pypi = SubmitField(word('PyPI'))
    github = SubmitField(word('GitHub'))
    cancel = SubmitField(word('Cancel'))
    delete = SubmitField(word('Delete'))


class GoogleDriveForm(FlaskForm):
    folder = SelectField(word('Folder'))
    submit = SubmitField(word('Save'))
    cancel = SubmitField(word('Cancel'))


class OneDriveForm(FlaskForm):
    folder = SelectField(word('Folder'))
    submit = SubmitField(word('Save'))
    cancel = SubmitField(word('Cancel'))


class GitHubForm(FlaskForm):
    shared = BooleanField(word('Access shared repositories'))
    orgs = BooleanField(word('Access organizational repositories'))
    save = SubmitField(word('Save changes'))
    configure = SubmitField(word('Configure'))
    unconfigure = SubmitField(word('Disable'))
    cancel = SubmitField(word('Back to profile'))


class TrainingForm(FlaskForm):
    the_package = HiddenField()
    the_file = HiddenField()
    the_group_id = HiddenField()
    show_all = HiddenField()
    submit = SubmitField(word('Save'))
    cancel = SubmitField(word('Cancel'))


class TrainingUploadForm(FlaskForm):
    usepackage = RadioField(word('Use Package'))
    jsonfile = FileField(word('JSON file'))
    importtype = RadioField(word('Import method'))
    submit = SubmitField(word('Import'))


class AddinUploadForm(FlaskForm):
    content = HiddenField()
    filename = HiddenField()


class FunctionFileForm(FlaskForm):
    pass


class APIKey(FlaskForm):
    action = HiddenField()
    key = HiddenField()
    security = HiddenField()
    name = StringField(word('Name'), validators=[validators.Length(min=1, max=255)])
    method = SelectField(word('Security Method'))
    permissions = SelectMultipleField(word('Limited Permissions'))
    submit = SubmitField(word('Create'))
    delete = SubmitField(word('Delete'))

    def validate(self, extra_validators=None):
        rv = FlaskForm.validate(self, extra_validators=extra_validators)
        if not rv:
            return False
        if self.action.data not in ('edit', 'new'):
            return False
        has_error = False
        if self.action.data in ('edit', 'new'):
            if (not isinstance(self.name.data, str)) or not re.search(r'[A-Za-z0-9]', self.name.data):
                self.name.errors.append(word("The name must be filled in."))
                has_error = True
            if (not isinstance(self.method.data, str)) or self.method.data not in ('referer', 'ip', 'none'):
                self.name.errors.append(word("You must select an option."))
                has_error = True
        if has_error:
            return False
        return True



import sys
import time
import psutil


def main():
    busy_pids = set()
    index = 0
    while True:
        busy_now = set()
        no_longer_busy = set()
        for pid in psutil.pids():
            try:
                p = psutil.Process(pid)
                if p.name() in ('apache2', 'uwsgi') and p.cpu_percent(interval=30.0) > 90.0:
                    busy_now.add(pid)
            except:
                continue
        index += 1
        index = index % 6
        if index == 5:
            for pid in busy_pids:
                if pid not in busy_now:
                    no_longer_busy.add(pid)
            for pid in no_longer_busy:
                busy_pids.discard(pid)
            for pid in busy_now:
                if pid in busy_pids:
                    try:
                        p = psutil.Process(pid)
                        p.kill()
                    except:
                        pass
                    sys.stderr.write("Killed " + str(pid) + "\n")
                    busy_pids.discard(pid)
                else:
                    busy_pids.add(pid)
        time.sleep(5)

if __name__ == '__main__':
    main()
